/*
 * ftpd/ls.c --
 *
 *	A very simple and basic ftp daemon. This file contains the
 *	directory listing code.
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "ftpd.h"

static int nflag = 1;

static void
show_mode(mode_t mode, FILE *stream)
{
    char c = '?';
    
    if (S_ISDIR(mode)) {
        c = 'd';
    } else if (S_ISREG(mode)) {
        c = '-';
    } else if (S_ISCHR(mode)) {
        c = 'c';
    } else if (S_ISBLK(mode)) {
        c = 'b';
#ifdef S_ISSOCK
    } else if (S_ISSOCK(mode)) {
        c = 's';
#endif
    } else if (S_ISFIFO(mode)) {
        c = 'f';
    }

    fputc(' ', stream);
    fputc(c, stream);
    fputc((mode & S_IRUSR) ? 'r' : '-', stream);
    fputc((mode & S_IWUSR) ? 'w' : '-', stream);
    fputc((mode & S_IXUSR) ? 'x' : '-', stream);
    fputc((mode & S_IRGRP) ? 'r' : '-', stream);
    fputc((mode & S_IWGRP) ? 'w' : '-', stream);
    fputc((mode & S_IXGRP) ? 'x' : '-', stream);
    fputc((mode & S_IROTH) ? 'r' : '-', stream);
    fputc((mode & S_IWOTH) ? 'w' : '-', stream);
    fputc((mode & S_IXOTH) ? 'x' : '-', stream);
}

static void
show_user(uid_t uid, FILE *stream)
{
    struct passwd *pwd;
    
    pwd = getpwuid(uid);
    if (!nflag && pwd) {
        fprintf(stream, " %8s", pwd->pw_name);
    } else {
        fprintf(stream, " %8u", (unsigned) uid);
    }
}

static void
show_group(gid_t gid, FILE *stream)
{
    struct group *grp;
    
    grp = getgrgid(gid);
    if (!nflag && grp) {
        fprintf(stream, " %8s", grp->gr_name);
    } else {
        fprintf(stream, " %8u", (unsigned) gid);
    }
}

static void
show_time(time_t t, time_t now, FILE *stream)
{
    struct tm *tm;
    char buffer[80];
    const time_t delta = 7905600;       /* 366/2 * 12 * 3600 seconds */
    size_t n = 0;

    tm = localtime(&t);
    if (now != (time_t) -1 && t < now - delta) {
        n = strftime(buffer, sizeof(buffer), "%b %d  %Y", tm);
    } else {
        n = strftime(buffer, sizeof(buffer), "%b %d %H:%M", tm);
    }
    if (n) {
        fprintf(stream, " %12s", buffer);
    } else {
        fprintf(stream, " %12lu", (unsigned long) t);
    }
}

void
ls(FILE *stream, const char *directory, int flags)
{
    DIR *d;
    struct dirent *e;
    time_t now;
    struct stat s;
    
    d = opendir(directory);
    if (! d) {
        syslog(LOG_WARNING, "opendir failed: %s", strerror(errno));
        return;
    }

    if (chdir(directory) == -1) {
        syslog(LOG_WARNING, "chdir failed: %s", strerror(errno));
        return;
    }

    now = time(NULL);
    
    while (1) {
        e = readdir(d);
        if (! e) {
            break;
        }
	if (e->d_name[0] == '.') continue;
	if (flags & FTPD_LS_LONG) {
	    if (stat(e->d_name, &s) == -1) {
		fprintf(stderr, "%s: %s: ", progname, e->d_name);
		syslog(LOG_WARNING, "stat failed: %s", strerror(errno));
		continue;
	    }
	    show_mode(s.st_mode, stream);
	    fprintf(stream, " %4d", s.st_nlink);
	    show_user(s.st_uid, stream);
	    show_group(s.st_gid, stream);
	    fprintf(stream, " %9llu", (unsigned long long) s.st_size);
	    show_time(s.st_mtime, now, stream);
	    fprintf(stream, " %s\r\n", e->d_name);
	} else {
	    fprintf(stream, "%s\r\n", e->d_name);
	}
    }

    closedir(d);
}

